﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Media;
using System.Net;
using System.Text;
using System.Windows.Forms;
using Encoder = System.Drawing.Imaging.Encoder;

namespace FlappyC3
{
    public partial class FlappyC3Game : Form
    {
        private int gravity_speed = 5;
        private int move_speed = 4;
        private int jump_height = 70;
        private int flying_speed = 7;
        private int column_distance = 400;
        private int forgiveness = 10;
        private Color deadColor = Color.Black;
        private int jump_goal;
        private readonly string flappy_folder;
        private readonly List<PictureBox> columns_a;
        private readonly List<PictureBox> columns_b;
        private bool isPlaying;
        private readonly int center;
        private const int floor = 400;
        private readonly Random randomizer;
        private int CurrentScore;
        private int HighScore;
        private bool isDead;
        private bool isPassingColumn;
        private const string ClientId = "fdf5f7630f95fc1"; //from imgur, specific to this program, do not use elsewhere BROKEN NEED NEW ID
        private readonly string config;
        private readonly string scorefile;
        public bool Mute = false;
        private SoundPlayer WavPlayer;
        private bool GotPoint;
        private int mouseX;
        private int mouseY;

        public FlappyC3Game()
        {
            SetStyle(ControlStyles.OptimizedDoubleBuffer | ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint, true);
            InitializeComponent();
            
            flappy_folder = Application.StartupPath + "\\res\\flappy\\";

            randomizer = new Random();
            columns_a = new List<PictureBox>{columnA1, columnA2, columnA3, columnA4, columnA5, columnA6};
            columns_b = new List<PictureBox> { columnB1, columnB2, columnB3, columnB4, columnB5, columnB6 };

            picBackground.Location = new Point(0, 0);
            picBackground.Size = Size;
            picBackground.SendToBack();

            WavPlayer = new SoundPlayer();

            center = (Height - flappy.Height)/2;
            lblRestart.Left = (Width - lblRestart.Width)/2;
            lblRestart.Parent = picBackground;
            lblCurrent.Left = (Width - lblCurrent.Width - lblCurrentScore.Width)/2;
            lblCurrent.Parent = picBackground;
            lblCurrentScore.Left = lblCurrent.Left + lblCurrent.Width;
            lblCurrentScore.Parent = picBackground;
            lblHigh.Left = (Width - lblHigh.Width - lblHighScore.Width)/2;
            lblHigh.Parent = picBackground;
            lblHighScore.Left = lblHigh.Left + lblHigh.Width;
            lblHighScore.Parent = picBackground;
            lblNew.Left = lblHigh.Left - lblNew.Width;
            lblNew.Parent = picBackground;
            lblScore.Left = (Width - lblScore.Width)/2;
            lblScore.Parent = picBackground;
            flappy.Top = center;
            flappy.Left = (Width - flappy.Width)/2;
            lblIntro.Left = (Width - lblIntro.Width)/2;
            lblIntro.Parent = picBackground;
            lblF1.Left = (Width - lblF1.Width)/2;
            lblF1.Parent = picBackground;
            picSpeaker.Parent = picBackground;

            config = flappy_folder + "flappy.config";
            scorefile = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + "\\c3\\fcs";

            if (File.Exists(config))
            {
                LoadConfig();
            }

            if (!File.Exists(scorefile)) return;
            var sr = new StreamReader(scorefile);
            try
            {
                HighScore = Convert.ToInt16(sr.ReadLine());
            }
            catch (Exception)
            {
                HighScore = 0;
            }
            sr.Dispose();
        }
        
        private void PlaySound(string sound)
        {
            var soundfile = flappy_folder + sound + ".wav";

            if (!File.Exists(soundfile) || Mute || (sound == "flap" && GotPoint)) return;
           
            WavPlayer = new SoundPlayer(soundfile);
            WavPlayer.Play();

            GotPoint = sound == "point";
            pointTmr.Enabled = GotPoint;
        }

        protected override CreateParams CreateParams
        {
            get
            {
                var cp = base.CreateParams;
                cp.ExStyle |= 0x02000000;  // Turn on WS_EX_COMPOSITED
                return cp;
            }
        } 

        private void LoadConfig()
        {
            if (!File.Exists(config)) return;
            
            var sr = new StreamReader(config);
            try
            {
                Text = GetConfigString(sr.ReadLine());
                //replace - to prevent cheating by using a negative value for gravity, etc
                gravity_speed = Convert.ToInt16(GetConfigString(sr.ReadLine().Replace("-","")));
                move_speed = Convert.ToInt16(GetConfigString(sr.ReadLine().Replace("-", "")));
                jump_height = Convert.ToInt16(GetConfigString(sr.ReadLine().Replace("-", "")));
                flying_speed = Convert.ToInt16(GetConfigString(sr.ReadLine().Replace("-", "")));
                column_distance = Convert.ToInt16(GetConfigString(sr.ReadLine().Replace("-", "")));
                forgiveness = Convert.ToInt16(GetConfigString(sr.ReadLine().Replace("-", "")));
                deadColor = ColorTranslator.FromHtml(GetConfigString(sr.ReadLine()));
            }
            catch (Exception)
            {
                MessageBox.Show("The configuration file was not formatted correctly and I deleted it\n\nIf you're going to modify this program follow the format of the one I will create when you close me down",
                    Text, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                sr.Dispose();
                DeleteFile(config);
                return;
            }
            sr.Dispose();
        }

        private void SaveConfig()
        {
            var sw = new StreamWriter(config, false);
            sw.WriteLine("GameName=" + Text);
            sw.WriteLine("GravitySpeed=" + gravity_speed);
            sw.WriteLine("HorizontalSpeed=" + move_speed);
            sw.WriteLine("JumpHeight=" + jump_height);
            sw.WriteLine("VerticalSpeed=" + flying_speed);
            sw.WriteLine("ColumnDistance=" + column_distance);
            sw.WriteLine("Forgiveness=" + forgiveness);
            sw.WriteLine("DeadColor=#" + deadColor.A.ToString("X2") + deadColor.R.ToString("X2") + deadColor.G.ToString("X2") + deadColor.B.ToString("X2"));
            sw.Dispose();
        }

        private void FlappyC3_KeyUp(object sender, KeyEventArgs e)
        {
             switch (e.KeyCode)
            {
                case Keys.Escape:
                    Close();
                    break;
                case Keys.F12:
                    GrabScreen();
                    break;    
                case Keys.F1:
                    MessageBox.Show("Flappy C3 'game' created by TrojanNemo\nWhy? Because!\n\nPress F2 for tutorial on how to customize\nPress F12 to take a screenshot\n\nEnjoy",
                        Text, MessageBoxButtons.OK, MessageBoxIcon.Information);    
                    break;
                case Keys.F2:
                    Process.Start("http://customscreators.com/index.php?/topic/10493-flappy-c3-the-game-how-to-play-customize/");
                    break;
                 case Keys.Space:
                    if (isDead) return;
                    doPlay();
                    break;
            }
        }

        private void GrabScreen()
        {
            var desktop = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);

            try
            {
                Bitmap bitmap;
                //capture image
                var location = Location;
                using (bitmap = new Bitmap(Width, Height))
                {
                    var g = Graphics.FromImage(bitmap);
                    g.CopyFromScreen(new Point(location.X, location.Y), Point.Empty, Size, CopyPixelOperation.SourceCopy);

                    //this is so image quality is higher than the default
                    var myEncoder = Encoder.Quality;
                    var myEncoderParameters = new EncoderParameters(1);
                    const long encoder = 100L;
                    var myEncoderParameter = new EncoderParameter(myEncoder, encoder);
                    myEncoderParameters.Param[0] = myEncoderParameter;
                    var myImageCodecInfo = GetEncoderInfo("image/png");

                    //prepare to prompt user for save location and extension
                    var sfd = new SaveFileDialog
                    {
                        Filter = "PNG Image|*.png",
                        Title = "Where should I save the screenshot to?",
                        FileName = "flappyc3_screenshot",
                        AddExtension = true,
                        InitialDirectory = desktop
                    };

                    if (sfd.ShowDialog() != DialogResult.OK || sfd.FileName == "") return;
                    DeleteFile(sfd.FileName);

                    //save image
                    bitmap.Save(sfd.FileName, myImageCodecInfo, myEncoderParameters);
                    

                    if (!File.Exists(sfd.FileName)) return;
                    MessageBox.Show("Screenshot saved successfully", "Success", MessageBoxButtons.OK,
                        MessageBoxIcon.Information);
                    /*if (MessageBox.Show("Screenshot saved successfully\nUpload to Imgur?", Text,
                                        MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
                    {
                        UploadToImgur(sfd.FileName);
                    }*/
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error saving screenshot\n'" + ex.Message + "'", Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public void UploadToImgur(string image)
        {
            var link = "";
            try
            {
                using (var w = new WebClient())
                {
                    var values = new NameValueCollection
                        {
                            {"image", Convert.ToBase64String(File.ReadAllBytes(image))}
                        };
                    w.Headers.Add("Authorization", "Client-ID " + ClientId);
                    var response = w.UploadValues("https://api.imgur.com/3/upload.xml", values);
                    var sr = new StreamReader(new MemoryStream(response), Encoding.Default);
                    while (sr.Peek() >= 0)
                    {
                        var line = sr.ReadLine();
                        if (line == null || !line.Contains("link")) continue;
                        var linkStart = line.IndexOf("<link>", StringComparison.Ordinal) + 6;
                        var linkEnd = line.IndexOf("</link>", linkStart, StringComparison.Ordinal);
                        link = line.Substring(linkStart, linkEnd - linkStart);
                    }
                    sr.Dispose();
                }
            }
            catch (Exception ex)
            {
                var error = ex.Message;
                link = "";
                if (error.Contains("429"))
                {
                    error = "Error Code 429: Rate limiting\nThis most likely means C3 users have uploaded too many images recently\nPlease wait a couple of hours and try again";
                }
                else if (error.Contains("500"))
                {
                    error = "Error Code 500: Unexpected internal error\nThis most likely means something is broken with the Imgur service\nPlease try again later";
                }
                if (MessageBox.Show("Sorry, there was an error uploading that image!\n\nThe error says:\n" + error + "\n\nTry again?", "Error",
                    MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == DialogResult.Yes)
                {
                    UploadToImgur(image);
                }
            }
            Process.Start(link);
        }
        
        private void gravity_Tick(object sender, EventArgs e)
        {
            if (flappy.Top + gravity_speed >= floor)
            {
                Lost();
                return;
            }
            flappy.Top = flappy.Top + gravity_speed;
        }

        private void Lost()
        {
            if (!isDead) //if lost by hitting ground, otherwise already done when hit the column
            {
                flappy.BackColor = deadColor;
                PlaySound("dead");
            }
            Cursor = Cursors.Default;
            isDead = true;
            gravity.Enabled = false;
            move.Enabled = false;
            isPlaying = false;
            flappy.Top = floor;
            lblNew.Visible = CurrentScore > HighScore;
            HighScore = CurrentScore > HighScore ? CurrentScore : HighScore;
            lblHighScore.Text = HighScore.ToString(CultureInfo.InvariantCulture);
            lblCurrentScore.Text = CurrentScore < 0 ? "0" : CurrentScore.ToString(CultureInfo.InvariantCulture);
            lblCurrent.Visible = true;
            lblCurrentScore.Visible = true;
            lblHigh.Visible = true;
            lblHighScore.Visible = true;
            lblRestart.Visible = true;
        }

        private void jump_Tick(object sender, EventArgs e)
        {
            if (flappy.Top - flying_speed <= 0 || flappy.Top - flying_speed <= jump_goal)
            {
                jump.Enabled = false;
                gravity.Enabled = true;
            }
            flappy.Top = flappy.Top - flying_speed;
        }

        private void HandleMouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left && isDead)
            {
                Cursor = Cursors.NoMove2D;
                mouseX = MousePosition.X;
                mouseY = MousePosition.Y;
            }
            if (isDead) return;
            doPlay();
        }

        private void doPlay()
        {
            if (!isPlaying)
            {
                lblF1.Visible = false;
                lblIntro.Visible = false;
                checkCollision.Enabled = true;
                move.Enabled = true;
            }

            PlaySound("flap");
            isPlaying = true;
            gravity.Enabled = false;
            jump_goal = flappy.Top - jump_height;
            jump.Enabled = true;
        }

        private void HideBoxes(IList<PictureBox> columns)
        {
            foreach (var column in columns)
            {
                column.Visible = true; // restore all of them
            }

            var hide = randomizer.Next(1, 5); //can't have the top-most or bottom-most be values
            columns[hide].Visible = false;
            if (hide + 1 >= 5)
            {
                columns[hide - 1].Visible = false;
            }
            else
            {
                columns[hide + 1].Visible = false;
            }
        }

        private void LoadGraphics()
        {
            try
            {
                //NemoLoadImage is a safer way to load the image into memory and let go of the resource
                //But it turns the image into a static bitmap, so for animated gifs we'll use the Image.FromFile instead
                picBackground.Image = File.Exists(flappy_folder + "bg.gif") ? Image.FromFile(flappy_folder + "bg.gif") : NemoLoadImage(flappy_folder + "bg.png");
                flappy.Image = File.Exists(flappy_folder + "flappy.gif") ? Image.FromFile(flappy_folder + "flappy.gif") : NemoLoadImage(flappy_folder + "flappy.png");

                flappy.BackColor = Color.Transparent;
                flappy.Parent = picBackground;

                columns_a[0].Image = File.Exists(flappy_folder + "purple.gif") ? Image.FromFile(flappy_folder + "purple.gif") : NemoLoadImage(flappy_folder + "purple.png");
                columns_a[1].Image = File.Exists(flappy_folder + "orange.gif") ? Image.FromFile(flappy_folder + "orange.gif") : NemoLoadImage(flappy_folder + "orange.png");
                columns_a[2].Image = File.Exists(flappy_folder + "blue.gif") ? Image.FromFile(flappy_folder + "blue.gif") : NemoLoadImage(flappy_folder + "blue.png");
                columns_a[3].Image = File.Exists(flappy_folder + "yellow.gif") ? Image.FromFile(flappy_folder + "yellow.gif") : NemoLoadImage(flappy_folder + "yellow.png");
                columns_a[4].Image = File.Exists(flappy_folder + "red.gif") ? Image.FromFile(flappy_folder + "red.gif") : NemoLoadImage(flappy_folder + "red.png");
                columns_a[5].Image = File.Exists(flappy_folder + "green.gif") ? Image.FromFile(flappy_folder + "green.gif") : NemoLoadImage(flappy_folder + "green.png");

                columns_b[0].Image = columns_a[0].Image;
                columns_b[1].Image = columns_a[1].Image;
                columns_b[2].Image = columns_a[2].Image;
                columns_b[3].Image = columns_a[3].Image;
                columns_b[4].Image = columns_a[4].Image;
                columns_b[5].Image = columns_a[5].Image;

                GoldStarA.Image = File.Exists(flappy_folder + "star.gif") ? Image.FromFile(flappy_folder + "star.gif") : NemoLoadImage(flappy_folder + "star.png");
                GoldStarA.BackColor = Color.Transparent;
                GoldStarA.Parent = picBackground;
                GoldStarB.Image = GoldStarA.Image;
                GoldStarB.BackColor = Color.Transparent;
                GoldStarB.Parent = picBackground;

                foreach (var column in columns_a)
                {
                    column.BackColor = Color.Transparent;
                    column.Parent = picBackground;
                }
                foreach (var column in columns_b)
                {
                    column.BackColor = Color.Transparent;
                    column.Parent = picBackground;
                }

                flappy.BringToFront();
                lblScore.BringToFront();
            }
            catch (Exception ex)
            {

                MessageBox.Show("Error loading image resources\n\n" + ex.Message, Text,MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private static void MoveColumn(IEnumerable<PictureBox> columns, int pos)
        {
            foreach (var column in columns)
            {
                column.Left = pos;
            }
        }

        private void FlappyC3_Shown(object sender, EventArgs e)
        {
            LoadConfig();
            LoadGraphics();
            ResetColumns();
        }

        private void ResetColumns()
        {
            MoveColumn(columns_a, Width);
            MoveColumn(columns_b, Width + column_distance);

            HideBoxes(columns_a);
            HideBoxes(columns_b);
        }

        private void move_Tick(object sender, EventArgs e)
        {
            MoveColumn(columns_a, columnA1.Left - move_speed);
            MoveColumn(columns_b, columnB1.Left - move_speed);
            GoldStarA.Left = GoldStarA.Left - move_speed;
            GoldStarB.Left = GoldStarB.Left - move_speed;

            if (GoldStarA.Left < 0 - GoldStarA.Width)
            {
                GoldStarA.Left = columnA1.Left + (column_distance - GoldStarA.Width)/2 + 15;
                MoveStar(GoldStarA);
            }
            else if (GoldStarB.Left < 0 - GoldStarB.Width)
            {
                GoldStarB.Left = columnB1.Left + (column_distance - GoldStarB.Width) / 2 + 15;
                MoveStar(GoldStarB);
            }
            if (columnA1.Left < 0 - columnA1.Width)
            {
                MoveColumn(columns_a, columnB1.Left + column_distance);
                HideBoxes(columns_a);
            }
            else if (columnB1.Left < 0 - columnB1.Width)
            {
                MoveColumn(columns_b, columnA1.Left + column_distance);
                HideBoxes(columns_b);
            }
        }

        private void MoveStar(Control star)
        {
            var bottom = Height - (GoldStarA.Height*2);
            var rnd = new Random();
            star.Top = rnd.Next(GoldStarA.Height, bottom);
            star.Visible = true;
        }

        private void showScore_Tick(object sender, EventArgs e)
        {
            lblScore.Visible = false;
            showScore.Enabled = false;
        }

        private void checkCollision_Tick(object sender, EventArgs e)
        {
            List<PictureBox> columns;

            var leftBorder = flappy.Left - GoldStarA.Width + forgiveness;
            var rightBorder = flappy.Left + flappy.Width - forgiveness;
            var topBorder = flappy.Top - GoldStarA.Height + forgiveness;
            var bottomBorder = flappy.Top + flappy.Height - forgiveness;

            var isGoldStar = false;
            var star = GoldStarA;
            if (GoldStarA.Left > leftBorder && GoldStarA.Left < rightBorder && GoldStarA.Top > topBorder &&
                GoldStarA.Top < bottomBorder && GoldStarA.Visible)
            {
                isGoldStar = true;
            }
            else if (GoldStarB.Left > leftBorder && GoldStarB.Left < rightBorder && GoldStarB.Top > topBorder &&
                GoldStarB.Top < bottomBorder && GoldStarB.Visible)
            {
                isGoldStar = true;
                star = GoldStarB;
            }
            if (isGoldStar)
            {
                DoScore();
                star.Visible = false;
            }
            
            leftBorder = flappy.Left - columnA1.Width + forgiveness;
            topBorder = flappy.Top - columnA1.Height + forgiveness;

            if (columnA1.Left > leftBorder && columnA1.Left < rightBorder)
            {
                isPassingColumn = true;
                columns = columns_a;
            }
            else if (columnB1.Left > leftBorder && columnB1.Left < rightBorder)
            {
                isPassingColumn = true;
                columns = columns_b;
            }
            else
            {
                if (!isPassingColumn) return;
                isPassingColumn = false;
                DoScore();
                return;
            }

            if (!columns.Any(column => column.Top > topBorder && column.Top < bottomBorder && column.Visible)) return;
            PlaySound("dead");
            isPassingColumn = false;
            move.Enabled = false;
            isDead = true;
            flappy.BackColor = deadColor;
            checkCollision.Enabled = false;
            GoldStarA.Visible = false;
            GoldStarB.Visible = false;
        }

        private void DoScore()
        {
            CurrentScore++;
            if (CurrentScore <= 0) return;
            PlaySound("point");
            lblScore.Text = CurrentScore.ToString(CultureInfo.InvariantCulture);
            lblScore.Visible = true;
            showScore.Enabled = true;
        }

        private void FlappyC3_FormClosing(object sender, FormClosingEventArgs e)
        {
            WavPlayer.Stop();

            var folder = Path.GetDirectoryName(scorefile);
            if (string.IsNullOrEmpty(folder)) return;

            if (!Directory.Exists(folder))
            {
                Directory.CreateDirectory(folder);
            }
            var sw = new StreamWriter(scorefile, false);
            sw.WriteLine(HighScore);
            sw.Dispose();

            SaveConfig();
        }

        private void pointTmr_Tick(object sender, EventArgs e)
        {
            GotPoint = false;
            pointTmr.Enabled = false;
        }

        private void picSpeaker_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.Left) return;
            Mute = !Mute;
            WavPlayer.Stop();
            picSpeaker.Image = Mute ? Properties.Resources.unmute : Properties.Resources.mute;
        }

        private void lblRestart_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.Left) return;
            lblCurrent.Visible = false;
            lblCurrentScore.Visible = false;
            lblHigh.Visible = false;
            lblHighScore.Visible = false;
            lblRestart.Visible = false;
            lblIntro.Visible = true;
            lblNew.Visible = false;
            lblF1.Visible = true;
            CurrentScore = 0;
            ResetColumns();
            flappy.BackColor = Color.Transparent;
            flappy.Top = center;
            isDead = false;
            GoldStarA.Visible = false;
            GoldStarB.Visible = false;
            WavPlayer.Stop();
        }

        private void picBackground_MouseUp(object sender, MouseEventArgs e)
        {
            Cursor = Cursors.Default;
        }

        private void picBackground_MouseMove(object sender, MouseEventArgs e)
        {
            if (Cursor != Cursors.NoMove2D) return;
            if (MousePosition.X != mouseX)
            {
                if (MousePosition.X > mouseX)
                {
                    Left = Left + (MousePosition.X - mouseX);
                }
                else if (MousePosition.X < mouseX)
                {
                    Left = Left - (mouseX - MousePosition.X);
                }
                mouseX = MousePosition.X;
            }

            if (MousePosition.Y == mouseY) return;
            if (MousePosition.Y > mouseY)
            {
                Top = Top + (MousePosition.Y - mouseY);
            }
            else if (MousePosition.Y < mouseY)
            {
                Top = Top - (mouseY - MousePosition.Y);
            }
            mouseY = MousePosition.Y;
        }

        #region Tools
        private static string GetConfigString(string raw_line)
        {
            var line = raw_line;
            var index = line.IndexOf("=", StringComparison.Ordinal) + 1;
            try
            {
                line = line.Substring(index, line.Length - index);
            }
            catch (Exception)
            {
                line = "";
            }
            return line.Trim();
        }

        private static void DeleteFile(string file)
        {
            if (file == "") return;
            if (!File.Exists(file)) return;
            try
            {
                File.Delete(file);
            }
            catch (Exception)
            { }
        }

        private static ImageCodecInfo GetEncoderInfo(String mimeType)
        {
            int j;
            var encoders = ImageCodecInfo.GetImageEncoders();
            for (j = 0; j < encoders.Length; ++j)
            {
                if (encoders[j].MimeType == mimeType)
                    return encoders[j];
            }
            return null;
        }

        private static Image NemoLoadImage(string file)
        {
            if (!File.Exists(file))
            {
                return null;
            }

            Image img;
            using (var bmpTemp = new Bitmap(file))
            {
                img = new Bitmap(bmpTemp);
            }

            return img;
        }
        #endregion
    }
}
